#include <string>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <errno.h>

#include "network.h"


using namespace std;

// Create connection from existing descriptor
Connection::Connection(int fd)
{
	this->fd = fd;
	async_recv_len = -1;
}

// Empty connection, which can connect with 'connect' method
Connection::Connection()
{
	this->fd = -1;
	async_recv_len = -1;
}

Connection::~Connection()
{
	if (fd != -1) {
		shutdown(fd,SHUT_RDWR);
		close(fd);
	};
}

int Connection::connect(string host, int port)
{
	struct sockaddr_in addr;

	if (inet_pton(AF_INET, host.c_str(), &addr.sin_addr.s_addr) == -1) {
		perror("inet_pton");
		return -1;
	};

	addr.sin_port = htons(port);
	addr.sin_family = AF_INET;
	
	fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd == -1) {
		perror("socket");
		return -1;
	};

	if (::connect(fd, (struct sockaddr*)&addr, sizeof(addr))) {
		perror("connect");
		return -1;
	};
	
	return 0;
}

int Connection::send(Packet &packet) 
{
	int ret;

	if (fd == -1) return -1;

	ret = ::send(fd, &packet.len, sizeof(packet.len), MSG_NOSIGNAL);
	if (ret == -1) {
		send_error(__FUNCTION__, __LINE__);
		return -1;
	};

	ret = ::send(fd, packet.payload, packet.len, MSG_NOSIGNAL);
	
	if (ret == packet.len) return 0;
	if (ret == -1) {
		send_error(__FUNCTION__, __LINE__);
		perror("send");
		return -1;
	};

	if (ret < packet.len) {
		fprintf(stderr, "Not all data were send; this is a BUG!!!\n");
		return -1;
	};

	return -2;
};

Packet Connection::receive()
{
	int ret, size;
	Packet data;

	data.len = 0;
	data.payload = NULL;

	// If there is async receive in progress, don't allow this receive
	if (async_recv_len != -1) return data;
	
	ret = recv(fd, &data.len, sizeof(size), MSG_WAITALL);
	if (ret == -1) 
		if (recv_error(__FUNCTION__, __LINE__)) {
			return data;
		} else {
			fprintf(stderr, "I'm too lazy to handle EAGAIN for length\n");
		};

		

	if (ret != sizeof(size)) {
		fprintf(stderr,"Something weird; I can't receive full length (ret is %d)\n", ret);
		return data;
	};
	
	data.payload = new char[data.len];
	size = data.len;
	while (size > 0) {
		ret = recv(fd, &((char*)data.payload)[data.len-size], size, MSG_WAITALL);
		if (ret == -1) {
			if (recv_error(__FUNCTION__, __LINE__)) {
				data.len = 0;
				delete static_cast<char*>(data.payload);
				data.payload = NULL;
				return data;
			};
			ret = 0;
		};

		size -= ret;
	};

	return data;
}

int Connection::receive_async(Packet &packet)
{
	int ret, len;

	if (async_recv_len == -1) {
		if (recv(fd, &len, sizeof(len), MSG_PEEK|MSG_DONTWAIT) != sizeof(len)) 
			return -1;
	
		// we'll assume that receive will get all data needed
		recv(fd, &async_recv_len, sizeof(len), MSG_WAITALL);
	};

	packet.payload = new char[async_recv_len];

	ret = recv(fd, packet.payload, async_recv_len, MSG_DONTWAIT|MSG_PEEK);
	if (ret == -1) {
		delete static_cast<char*>(packet.payload); packet.payload = NULL;
		// no data available at the moment
		if (recv_error(__FUNCTION__, __LINE__)) {
			shutdown(fd, SHUT_RDWR);
			close(fd);
			fd = -1;
			
			return -1;
		}

		return -1;
	};

	if (ret != async_recv_len) {
		delete static_cast<char*>(packet.payload); packet.payload = NULL;
		return -1;
	};

	ret = recv(fd, packet.payload, async_recv_len, MSG_DONTWAIT);
	if (ret == -1) {
		recv_error(__FUNCTION__, __LINE__);
		delete static_cast<char*>(packet.payload); packet.payload = NULL;
		return -1;
	};

	// packet successfully received
	packet.len = async_recv_len;
	async_recv_len = -1;

	return 0;
}

bool Connection::good()
{
	struct pollfd p;

	if (fd == -1) return -1;

	p.fd = fd;
	p.events = POLLERR|POLLHUP|POLLNVAL;

	poll(&p, 1, 0);
	
	if (p.revents > 0) {
		/* Error condition - clean up and make connection invalid */
		close(fd);
		fd = -1;
	};
	
	return fd > -1;
}

int Connection::recv_error(const char *fun, int line)
{
	if (errno == EAGAIN) return 0;
	fprintf(stderr, "%s(%d): %s\n", fun, line, strerror(errno) );
	return 1;
};

int Connection::send_error(const char *fun, int line)
{
	if (errno == EAGAIN) return 0;
	fprintf(stderr, "%s(%d): %s\n", fun, line, strerror(errno) );
	return 1;
};


/*
 * class NetworkServer
 */


NetworkServer::NetworkServer(string interface, int port)
{
	struct sockaddr_in addr;

	// Create socket
	fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd == -1) perror("socket");

	// Bind it to specified port
	inet_pton(AF_INET, interface.c_str(), &addr.sin_addr.s_addr);
	addr.sin_port = htons(port);
	addr.sin_family = AF_INET;

	if (bind( fd, (struct sockaddr *) &addr, sizeof(addr)) == -1 ) perror("bind");

	// Set it to listen state, backlog for 10 conns should be enough
	listen(fd, 10);
}

NetworkServer::~NetworkServer()
{
	close(fd);
}

bool NetworkServer::connection_pending()
{
	struct pollfd p;

	p.fd = fd;
	// Which one is on when connection is pending?
	p.events = POLLIN | POLLPRI;
	p.revents = 0;

	poll(&p, 1, 0);
	// TODO: implement
	return p.revents > 0;
}

Connection * NetworkServer::accept_connection()
{
	int conn = accept(fd, NULL, NULL);

	if (conn == -1) {
		perror("accept");
		return NULL;
	};

	return new Connection (conn);
}


